/************************************************************************
 *   IRC - Internet Relay Chat, doc/example_module.c
 *   Copyright (C) 2001 Hybrid Development Team
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   $Id: m_svs.c,v 1.18.2.4 2004/04/11 00:20:48 jeremy Exp $
 */

/* List of ircd includes from ../include/ */
#include "stdinc.h"
#include "handlers.h"
#include "client.h"
#include "common.h"     /* FALSE bleah */
#include "ircd.h"
#include "irc_string.h"
#include "numeric.h"
#include "fdlist.h"
#include "s_bsd.h"
#include "s_conf.h"
#include "s_log.h"
#include "s_serv.h"
#include "send.h"
#include "msg.h"
#include "parse.h"
#include "modules.h"

#include "s_user.h"  /* send_umode_out() */
#include "channel_mode.h" /* MODE_ADD & MOD_DEL */
#include "hash.h"

static int clean_nick_name(char *);
static int clean_user_name(char *);
static int clean_host_name(char *);
static void change_nick (struct Client *, struct Client *, struct Client *, char const *,
			char const *);

/* SVS* */
static void mo_svsnick	(struct Client *, struct Client *, int, char **);
static void mo_svshost	(struct Client *, struct Client *, int, char **);
static void mo_svsmode	(struct Client *, struct Client *, int, char **);
static void mo_svsident	(struct Client *, struct Client *, int, char **);

static void ms_svsnick  (struct Client *, struct Client *, int, char **);
static void ms_svshost  (struct Client *, struct Client *, int, char **);
static void ms_svsident (struct Client *, struct Client *, int, char **);

struct Message svshost_msgtab = { "SVSHOST", 0, 0, 3, 3, MFLG_SLOW, 0, { m_ignore, m_ignore, ms_svshost, m_ignore } };
struct Message svsnick_msgtab = { "SVSNICK", 0, 0, 4, 4, MFLG_SLOW, 0, { m_ignore, m_ignore, ms_svsnick, m_ignore } };
struct Message svsmode_msgtab = { "SVSMODE", 0, 0, 3, 3, MFLG_SLOW, 0, { m_ignore, m_ignore, mo_svsmode, m_ignore } };
struct Message svsident_msgtab = { "SVSIDENT", 0, 0, 3, 3, MFLG_SLOW, 0, { m_ignore, m_ignore, ms_svsident, m_ignore } };

struct Message chghost_msgtab = { "CHGHOST", 0, 0, 3, 3, MFLG_SLOW, 0, { m_ignore, m_not_oper, m_ignore, mo_svshost } };
struct Message chgnick_msgtab = { "CHGNICK", 0, 0, 3, 3, MFLG_SLOW, 0, { m_ignore, m_not_oper, m_ignore, mo_svsnick } };
struct Message chgmode_msgtab = { "CHGMODE", 0, 0, 3, 3, MFLG_SLOW, 0, { m_ignore, m_not_oper, m_ignore, mo_svsmode } };
struct Message chgident_msgtab = { "CHGIDENT", 0, 0, 3, 3, MFLG_SLOW, 0, { m_ignore, m_not_oper, m_ignore, mo_svsident } };

struct Message sethost_msgtab = { "SETHOST", 0, 0, 3, 3, MFLG_SLOW, 0, { m_ignore, m_not_oper, m_ignore, mo_svshost } };
struct Message setnick_msgtab = { "SETNICK", 0, 0, 3, 3, MFLG_SLOW, 0, { m_ignore, m_not_oper, m_ignore, mo_svsnick } };
struct Message setmode_msgtab = { "SETMODE", 0, 0, 3, 3, MFLG_SLOW, 0, { m_ignore, m_not_oper, m_ignore, mo_svsmode } };
struct Message setident_msgtab = { "SETIDENT", 0, 0, 3, 3, MFLG_SLOW, 0, { m_ignore, m_not_oper, m_ignore, mo_svsident } };

#ifndef STATIC_MODULES
void _modinit (void)
{
	mod_add_cmd (&svshost_msgtab);
	mod_add_cmd (&svsnick_msgtab);
	mod_add_cmd (&svsmode_msgtab);
	mod_add_cmd (&svsident_msgtab);

	mod_add_cmd (&chghost_msgtab);
	mod_add_cmd (&chgnick_msgtab);
	mod_add_cmd (&chgmode_msgtab);
	mod_add_cmd (&chgident_msgtab);

	mod_add_cmd (&sethost_msgtab);
	mod_add_cmd (&setnick_msgtab);
	mod_add_cmd (&setmode_msgtab);
	mod_add_cmd (&setident_msgtab);
}

void _moddeinit (void)
{
    mod_del_cmd (&svshost_msgtab);
    mod_del_cmd (&svsnick_msgtab);
    mod_del_cmd (&svsmode_msgtab);
    mod_del_cmd (&svsident_msgtab);

    mod_del_cmd (&chghost_msgtab);
    mod_del_cmd (&chgnick_msgtab);
    mod_del_cmd (&chgmode_msgtab);
    mod_del_cmd (&chgident_msgtab);

    mod_del_cmd (&sethost_msgtab);
    mod_del_cmd (&setnick_msgtab);
    mod_del_cmd (&setmode_msgtab);
    mod_del_cmd (&setident_msgtab);
}

const char *_version = "$Revision: 1.18.2.4 $";
#endif

static void mo_svsnick (struct Client *client_p, struct Client *source_p, int parc, char *parv[])
{
	struct Client *target_p, *nickname_p;
	char target[NICKLEN], nickname[NICKLEN];

	if (ConfigFileEntry.disable_svs_commands)
	{
		sendto_one (source_p, ":%s NOTICE %s :SVS Commands disabled.", me.name, source_p->name);
		return;
	}

	strlcpy (target, parv[1], NICKLEN);
	strlcpy (nickname, parv[2], NICKLEN);

	if (!strcmp(nickname, target))
		return;

	if ((target_p = find_client (target)) != NULL)
	{
		if (!clean_nick_name(nickname))
		{
			sendto_one(source_p, form_str(ERR_ERRONEUSNICKNAME), me.name, source_p->name, nickname);
			return;
		}

		if (find_nick_resv(nickname))
		{
			sendto_one(source_p, form_str(ERR_UNAVAILRESOURCE), me.name, source_p->name, nickname);
			return;
		}

		if (!strcasecmp(target, nickname))
		{
			change_nick (source_p, target_p, client_p, (char const *)target,
					(const char *)nickname);
			return;
		}

		if ((nickname_p = find_client (nickname)) != NULL)
		{
			sendto_one(source_p, form_str(ERR_NICKNAMEINUSE), me.name, source_p->name, nickname);
			return;
		}

		change_nick (source_p, target_p, client_p, (char const *)target,
				(const char *)nickname);

		return;
	}

	else
	{
		sendto_one(source_p, form_str(ERR_NOSUCHNICK), me.name, source_p->name, nickname);
		return;
	}

	/* NOT REACHED */
	return;
}

static void mo_svshost (struct Client *client_p, struct Client *source_p, int parc, char *parv[])
{
	struct Client *target_p;
	char target[NICKLEN], hostname[HOSTLEN];
	int setflags;

	if (ConfigFileEntry.disable_svs_commands)
	{
		sendto_one (source_p, ":%s NOTICE %s :SVS Commands disabled.", me.name, source_p->name);
		return;
	}

	strlcpy (target, parv[1], NICKLEN+1);
	strlcpy (hostname, parv[2], HOSTLEN+1);

	if ((target_p = find_client (target)) != NULL)
	{
		if (!valid_hostname (hostname) || hostname == NULL)
		{
			sendto_one (source_p, ":%s NOTICE %s :Invalid hostname", me.name, source_p->name);
			return;
		}

		if (!strchr (hostname, '.') && !strchr (hostname, ':'))
		{
			sendto_one (source_p, ":%s NOTICE %s :Invalid hostname", me.name, source_p->name);
			return;
		}

		if (!strcmp (target_p->host, hostname))
			return;

		if (MyClient(target_p) && !IsMaskHost(target_p))
		{
			setflags = target_p->umodes;
			target_p->umodes |= UMODE_MASKHOST;
			send_umode_out (target_p, target_p, setflags);
		}

		strlcpy (target_p->host, hostname, HOSTLEN+1);
		sendto_one (source_p, ":%s NOTICE %s :Changed %s's hostname to %s", me.name, source_p->name, target_p->name, hostname);
		sendto_server (NULL, NULL, NOCAPS, NOCAPS, ":%s SVSHOST %s %s", me.name, target_p->name, target_p->host);
        sendto_realops_flags (UMODE_SPY, L_ALL, "SVSHOST called for %s[%s] (%s)", target, hostname, get_oper_name(source_p));
        sendto_server (client_p, NULL, NOCAPS, NOCAPS, ":%s ENCAP * OPERSPY SVSHOST :called for %s[%s]", source_p->name, target, hostname);
		return;
	}

	sendto_one(source_p, form_str(ERR_NOSUCHNICK), me.name, source_p->name, target);
	return;
}

static void mo_svsmode (struct Client *client_p, struct Client *source_p, int parc, char *parv[])
{
	struct Client *target_p;
	char target[NICKLEN], **p, *m;
	int flag, i, what, setflags;

	if (ConfigFileEntry.disable_svs_commands)
	{
		sendto_one (source_p, ":%s NOTICE %s :SVS Commands disabled.", me.name, source_p->name);
		return;
	}

	strlcpy (target, parv[1], NICKLEN);

	what = MODE_ADD;

	if ((target_p = find_client (target)) != NULL)
	{
		setflags = target_p->umodes;

		for (p = &parv[2]; p && *p; p++)
		{
			for (m = *p; *m; m++)
			{
				switch (*m)
				{
					case '+':
						what = MODE_ADD;
						break;
					case '-':
						what = MODE_DEL;
						break;
					case 'o':
						break;
					case ' ':
					case '\n':
					case '\r':
					case '\t':
						break;

					default:
						if ((flag = user_modes_from_c_to_bitmask[(unsigned char)*m]))
						{
							if ((ConfigFileEntry.oper_only_umodes & flag) && (!IsOper(target_p)))
								;
							else
							{
								if (what == MODE_ADD)
									target_p->umodes |= flag;
								else
									target_p->umodes &= ~flag;
							}
						}
						break;
				}
			}
		}

		send_umode_out (target_p, target_p, setflags);
        sendto_realops_flags (UMODE_SPY, L_ALL, "SVSMODE called for %s[%s] (%s)", target, parv[2], get_oper_name(source_p));
        sendto_server (client_p, NULL, CAP_ENCAP, NOCAPS, ":%s ENCAP * OPERSPY SVSMODE called for %s[%s]", source_p->name, target, parv[2]);
		return;
	}

	sendto_one(source_p, form_str(ERR_NOSUCHNICK), me.name, source_p->name, target);
	return;
}

static void mo_svsident (struct Client *client_p, struct Client *source_p, int parc, char *parv[])
{
	struct Client *target_p;
	char target[NICKLEN], username[USERLEN];

	if (ConfigFileEntry.disable_svs_commands)
	{
		sendto_one (source_p, ":%s NOTICE %s :SVS Commands disabled.", me.name, source_p->name);
		return;
	}

	strlcpy (target, parv[1], NICKLEN);
	strlcpy (username, parv[2], USERLEN);

	if ((target_p = find_client (target)) != NULL)
	{
		if (!clean_user_name(username))
		{
			sendto_one (source_p, ":%s NOTICE %s :Invalid username", me.name, source_p->name);
			return;
		}

		if (!strcmp (target_p->username, username))
			return;

		strlcpy (target_p->username, username, USERLEN);
		sendto_one (source_p, ":%s NOTICE %s :Changed %s's ident to %s", me.name, source_p->name, target, username);
		sendto_server (NULL, NULL, NOCAPS, NOCAPS, ":%s SVSIDENT %s %s", me.name, target, username);
		sendto_server (NULL, NULL, CAP_ENCAP, NOCAPS, ":%s ENCAP * OPERSPY SVSIDENT called for %s[%s]", source_p->name, target, username);
		return;
	}

	sendto_one(source_p, form_str(ERR_NOSUCHNICK), me.name, source_p->name, target);
	return;
}

static void ms_svshost (struct Client *client_p, struct Client *source_p, int parc, char *parv[])
{
	struct Client *target_p;
	char target[NICKLEN], hostname[HOSTLEN];
	int setflags;

	strlcpy (target, parv[1], NICKLEN);
	strlcpy (hostname, parv[2], HOSTLEN);

	if (!IsServer(source_p))
		return;

	target_p = find_client (target);
	strlcpy (target_p->host, hostname, HOSTLEN);

	if (MyClient(target_p) && !IsMaskHost(target_p))
	{
		setflags = target_p->umodes;
		target_p->umodes |= UMODE_MASKHOST;
		send_umode_out (target_p, target_p, setflags);
	}

	sendto_server (client_p, NULL, NOCAPS, NOCAPS, ":%s SVSHOST %s %s", me.name, target, hostname);
	return;
}

static void ms_svsnick (struct Client *client_p, struct Client *source_p, int parc, char *parv[])
{
	struct Client *target_p;
	char nickname[NICKLEN], target[NICKLEN];

	strlcpy (nickname, parv[2], NICKLEN);
	strlcpy (target, parv[1], NICKLEN);

	target_p = find_client (target);

	target_p->tsinfo = atoi(parv[3]);

	if (MyClient(target_p))
		sendto_realops_flags (UMODE_NCHANGE, L_ALL, "Nick change: From %s to %s [%s@%s]", target_p->name, nickname, target_p->username, target_p->host);

	sendto_common_channels_local (target_p, ":%s!%s@%s NICK :%s", target_p->name, target_p->username, target_p->host, nickname);

	if (target_p->user)
		add_history (target_p, 1);

	del_from_client_hash_table (target_p->name, target_p);
	strcpy (target_p->name, nickname);
	add_to_client_hash_table (nickname, target_p);

	del_all_accepts (target_p);

	if (MyClient(target_p))
		fd_note (target_p->localClient->fd, "Nick: %s", nickname);

	sendto_server (client_p, NULL, NOCAPS, NOCAPS, ":%s SVSNICK %s %s %lu", me.name, target, nickname, target_p->tsinfo);
	return;
}

static void ms_svsident (struct Client *client_p, struct Client *source_p, int parc, char *parv[])
{
	struct Client *target_p;
	char target[NICKLEN], username[USERLEN];

	if (!IsServer(source_p))
		return;

	strlcpy (target, parv[1], NICKLEN);
	strlcpy (username, parv[2], USERLEN);

	target_p = find_client (target);

	strlcpy (target_p->username, username, USERLEN);
	sendto_server (client_p, NULL, NOCAPS, NOCAPS, ":%s SVSIDENT %s %s", me.name, target, username);
	return;
}

/* clean_nick_name()
 *
 * input        - nickname
 * output       - none
 * side effects - walks through the nickname, returning 0 if erroneous
 */
static int
clean_nick_name(char *nick)
{
        s_assert(nick);
        if(nick == NULL)
                return 0;

        /* nicks cant start with a digit or -, and must have a length */
        if(*nick == '-' || IsDigit(*nick) || !*nick)
                return 0;

        for (; *nick; nick++)
        {
                if(!IsNickChar(*nick))
                        return 0;
        }

        return 1;
}

/* clean_user_name()
 *
 * input        - username
 * output       - none
 * side effects - walks through the username, returning 0 if erroneous
 */
static int
clean_user_name(char *user)
{
        s_assert(user);
        if(user == NULL)
                return 0;

        for (; *user; user++)
        {
                if(!IsUserChar(*user))
                        return 0;

        }

        return 1;
}

/* clean_host_name()
 * input        - hostname
 * output       - none
 * side effects - walks through the hostname, returning 0 if erroneous
 */
static int
clean_host_name(char *host)
{
        s_assert(host);
        if(host == NULL)
                return 0;
        for (; *host; host++)
        {
                if(!IsHostChar(*host))
                        return 0;
        }

        return 1;
}

static void change_nick (struct Client *source_p, struct Client *target_p,
			struct Client *client_p, char const *old, char const *new)
{
    target_p->tsinfo = CurrentTime;

    if (MyClient(target_p))
        sendto_realops_flags (UMODE_NCHANGE, L_ALL,
			"Nick change: From %s to %s [%s@%s]",
			target_p->name, new, target_p->username, target_p->host);

    sendto_common_channels_local (target_p, ":%s!%s@%s NICK :%s", target_p->name,
				target_p->username, target_p->host, new);

    if (target_p->user)
        add_history (target_p, 1);

    del_from_client_hash_table (target_p->name, target_p);
    strcpy (target_p->name, new);
    add_to_client_hash_table (new, target_p);

    del_all_accepts (target_p);

    if (MyClient(target_p))
        fd_note (target_p->localClient->fd, "Nick: %s", new);

    sendto_one (source_p, ":%s NOTICE %s :Changed %s's nickname to %s", me.name, source_p->name,
		old, new);

    sendto_server (client_p, NULL, NOCAPS, NOCAPS, ":%s SVSNICK %s %s %lu", me.name, old, new,
		target_p->tsinfo);

    sendto_realops_flags (UMODE_SPY, L_ALL, "SVSNICK called for %s[%s] [%s]", old,
		new, get_oper_name(source_p));

    sendto_server (client_p, NULL, NOCAPS, NOCAPS, ":%s ENCAP * OPERSPY SVSNICK called"
		" for %s[%s] [%s]", source_p->name, old, new);

    return;
}
